------------------------------------------------------------------------
-- Equality of types and terms is decidable
------------------------------------------------------------------------

-- # import Explicit
-- # import Implicit
-- # import Normalisation
-- # import Eq
-- # import TmVar

infix 42 ≟, ≜?

------------------------------------------------------------------------
-- Types for possible equality

data CtxtEq? (Γ¹, Γ² : Ctxt) : Set where
  equalCtxt    : Γ¹ ≜ Γ² -> CtxtEq? Γ¹ Γ²
  notEqualCtxt : CtxtEq? Γ¹ Γ²

data TyEq? {Γ¹, Γ² : Ctxt} (τ¹ : Ty Γ¹) (τ² : Ty Γ²) : Set where
  equalTy    : τ¹ ≛ τ² -> TyEq? τ¹ τ²
  notEqualTy : TyEq? τ¹ τ²

data VarEq? {Γ¹, Γ² : Ctxt} {τ¹ : Ty Γ¹} {τ² : Ty Γ²}
            (v¹ : Γ¹ ∋ τ¹) (v² : Γ² ∋ τ²) : Set where
  equalVar    : v¹ ≈∋ v² -> VarEq? v¹ v²
  notEqualVar : VarEq? v¹ v²

data TmEq? {Γ¹, Γ² : Ctxt} {τ¹ : Ty Γ¹} {τ² : Ty Γ²}
           (t¹ : Γ¹ ⊢ τ¹) (t² : Γ² ⊢ τ²) : Set where
  equalTm    : t¹ ≈⊢ t² -> TmEq? t¹ t²
  notEqualTm : TmEq? t¹ t²

data Tm⁻Eq? {Γ¹ : Ctxt} {τ¹ : Ty Γ¹} {t¹ : Γ¹ ⊢ τ¹} {tv¹ : TmVar}
            {Γ² : Ctxt} {τ² : Ty Γ²} {t² : Γ² ⊢ τ²} {tv² : TmVar}
            (t⁻¹ : Tm⁻ tv¹ t¹) (t⁻² : Tm⁻ tv² t²) : Set where
  equalTm⁻    : t⁻¹ ≈⊢⁻ t⁻² -> Tm⁻Eq? t⁻¹ t⁻²
  notEqualTm⁻ : Tm⁻Eq? t⁻¹ t⁻²

data NFEq? {Γ¹ : Ctxt} {τ¹ : Ty Γ¹} {t¹ : Γ¹ ⊢ τ¹}
           {Γ² : Ctxt} {τ² : Ty Γ²} {t² : Γ² ⊢ τ²}
           (nf¹ : NF t¹) (nf² : NF t²) : Set where
  equalNF    : nf¹ ≈⊨ nf² -> NFEq? nf¹ nf²
  notEqualNF : NFEq? nf¹ nf²

data AtomEq? {Γ¹ : Ctxt} {τ¹ : Ty Γ¹} {t¹ : Γ¹ ⊢ τ¹}
             {Γ² : Ctxt} {τ² : Ty Γ²} {t² : Γ² ⊢ τ²}
             (at¹ : Atom t¹) (at² : Atom t²) : Set where
  equalAtom    : at¹ ≈⊨′ at² -> AtomEq? at¹ at²
  notEqualAtom : AtomEq? at¹ at²

------------------------------------------------------------------------
-- Decision procedures for equality

equalVar? (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
          (Γ² : Ctxt) (τ² : Ty Γ²)
  : Γ¹ ≜ Γ² -> (v¹ : Γ¹ ∋ τ¹) -> (v² : Γ² ∋ τ²) -> VarEq? v¹ v²
equalVar? (eqΓ ▻Cong eqσ) (vz σ¹)    (vz σ²)    = equalVar (vzCong eqσ)
equalVar? (eqΓ ▻Cong eqσ) (vs v¹ σ¹) (vs v² σ²)
  with equalVar? eqΓ v¹ v² =>
  _ _               _          _          notEqualVar    = notEqualVar
  _ (eqΓ ▻Cong eqσ) (vs v¹ σ¹) (vs v² σ²) (equalVar eqv) =
    equalVar (vsCong eqv eqσ)
equalVar? eqΓ (v¹ ∷≡∋ eqτ) v² with equalVar? eqΓ v¹ v² =>
  _ _   _            _  notEqualVar    = notEqualVar
  _ eqΓ (v¹ ∷≡∋ eqτ) v² (equalVar eqv) = equalVar (castVarEqˡ eqτ eqv)
equalVar? eqΓ v¹ (v² ∷≡∋ eqτ) with equalVar? eqΓ v¹ v² =>
  _ _   _  _            notEqualVar    = notEqualVar
  _ eqΓ v¹ (v² ∷≡∋ eqτ) (equalVar eqv) = equalVar (castVarEqʳ eqτ eqv)
equalVar? _ _ _ = notEqualVar

mutual

  equalNF? (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
           (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : τ¹ ≛ τ² -> (nf¹ : NF t¹) -> (nf² : NF t²) -> NFEq? nf¹ nf²
  equalNF? eqτ (✶Atom at¹) (✶Atom at²)
    with equalAtom? (tyEqCtxtEq eqτ) at¹ at² =>
    equalNF?₁ eqτ (✶Atom at¹) (✶Atom at²) (equalAtom eqat) =
      equalNF (✶AtomCong eqat)
    equalNF?₁ _ _ _ notEqualAtom = notEqualNF
  equalNF? eqτ (ElAtom at¹) (ElAtom at²)
    with equalAtom? (tyEqCtxtEq eqτ) at¹ at² =>
    equalNF?₂ eqτ (ElAtom at¹) (ElAtom at²) (equalAtom eqat) =
      equalNF (ElAtomCong eqat)
    equalNF?₂ _ _ _ notEqualAtom = notEqualNF
  equalNF? (ΠCong eqτ₁ eqτ₂) (λ⊨ nf¹) (λ⊨ nf²)
    with equalNF? eqτ₂ nf¹ nf² =>
    equalNF?₃ (ΠCong eqτ₁ eqτ₂) (λ⊨ nf¹) (λ⊨ nf²) (equalNF eqnf) =
      equalNF (λ⊨Cong eqnf)
    equalNF?₃ _ _ _ notEqualNF = notEqualNF
  equalNF? eqτ (nf¹ ∷≡⊨ eqt) nf²
    with equalNF? (tmEqTyEq eqt `transTy` eqτ) nf¹ nf² =>
    equalNF?₄ eqτ (nf¹ ∷≡⊨ eqt) nf² (equalNF eqnf) =
      equalNF (castNFEqˡ eqt eqnf)
    equalNF?₄ _ _ _ notEqualNF = notEqualNF
  equalNF? eqτ nf¹ (nf² ∷≡⊨ eqt)
    with equalNF? (eqτ `transTy` symTy (tmEqTyEq eqt)) nf¹ nf² =>
    equalNF?₅ eqτ nf¹ (nf² ∷≡⊨ eqt) (equalNF eqnf) =
      equalNF (castNFEqʳ eqt eqnf)
    equalNF?₅ _ _ _ notEqualNF = notEqualNF
  equalNF? _ _ _ = notEqualNF

  equalAtom? (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : Γ¹ ≜ Γ² -> (at¹ : Atom t¹) -> (at² : Atom t²) -> AtomEq? at¹ at²
  equalAtom? eqΓ (var⊨ v¹) (var⊨ v²) with equalVar? eqΓ v¹ v² =>
    _ _   _         _         notEqualVar    = notEqualAtom
    _ eqΓ (var⊨ v¹) (var⊨ v²) (equalVar eqv) = equalAtom (var⊨Cong eqv)
  equalAtom? eqΓ (at₁¹ @⊨ nf₂¹) (at₁² @⊨ nf₂²)
    with equalAtom? eqΓ at₁¹ at₁² =>
    _ _   _              _              notEqualAtom      = notEqualAtom
    _ eqΓ (at₁¹ @⊨ nf₂¹) (at₁² @⊨ nf₂²) (equalAtom eqat₁)
      with equalNF? (ΠEqLeftEq $ tmEqTyEq (atomEqTmEq eqat₁))
           nf₂¹ nf₂² =>
      _ _ _ _ _ notEqualNF = notEqualAtom
      _ eqΓ (at₁¹ @⊨ nf₂¹) (at₁² @⊨ nf₂²)
        (equalAtom eqat₁) (equalNF eqnf₂) =
        equalAtom (app⊨Cong eqat₁ eqnf₂)
  equalAtom? eqΓ (at¹ ∷≡⊨′ eqt) at²
    with equalAtom? eqΓ at¹ at² =>
    _ _ _ _ notEqualAtom = notEqualAtom
    _ eqΓ (at¹ ∷≡⊨′ eqt) at² (equalAtom eqat) =
      equalAtom (castAtomEqˡ eqt eqat)
  equalAtom? eqΓ at¹ (at² ∷≡⊨′ eqt)
    with equalAtom? eqΓ at¹ at² =>
    _ _ _ _ notEqualAtom = notEqualAtom
    _ eqΓ at¹ (at² ∷≡⊨′ eqt) (equalAtom eqat) =
      equalAtom (castAtomEqʳ eqt eqat)
  equalAtom? _ _ _ = notEqualAtom

equalTm⁻? (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
          (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
  :  τ¹ ≛ τ² -> (t⁻¹ : Tm⁻ tv¹ t¹) -> (t⁻² : Tm⁻ tv² t²)
  -> Tm⁻Eq? t⁻¹ t⁻²
equalTm⁻? eqτ t⁻¹ t⁻² with equalNF? eqτ (eval t⁻¹) (eval t⁻²) =>
  equalTm⁻?₁ _ _ _ (equalNF eqnf) = equalTm⁻ (tmEqTm⁻Eq $ nfEqTmEq eqnf)
  equalTm⁻?₁ _ _ _ notEqualNF     = notEqualTm⁻

equalTm? (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
         (Γ² : Ctxt) (τ² : Ty Γ²)
  : τ¹ ≛ τ² -> (t¹ : Γ¹ ⊢ τ¹) -> (t² : Γ² ⊢ τ²) -> TmEq? t¹ t²
equalTm? eqτ t¹ t² with equalTm⁻? eqτ (tmToTm⁻ t¹) (tmToTm⁻ t²) =>
  _ _ _ _ (equalTm⁻ eqt⁻) = equalTm (tm⁻EqTmEq eqt⁻)
  _ _ _ _ notEqualTm⁻     = notEqualTm

equalTy? (Γ¹, Γ² : Ctxt)
  : Γ¹ ≜ Γ² -> (τ¹ : Ty Γ¹) -> (τ² : Ty Γ²) -> TyEq? τ¹ τ²
equalTy? eqΓ (✶ Γ¹)      (✶ Γ²)      = equalTy (✶Cong eqΓ)
equalTy? eqΓ (Π τ₁¹ τ₂¹) (Π τ₁² τ₂²) with equalTy? eqΓ τ₁¹ τ₁² =>
  _ _   _           _           notEqualTy     = notEqualTy
  _ eqΓ (Π τ₁¹ τ₂¹) (Π τ₁² τ₂²) (equalTy eqτ₁)
    with equalTy? (eqΓ ▻Cong eqτ₁) τ₂¹ τ₂² =>
    _ _ _ _ _ notEqualTy = notEqualTy
    _ eqΓ (Π τ₁¹ τ₂¹) (Π τ₁² τ₂²) (equalTy eqτ₁) (equalTy eqτ₂) =
      equalTy (ΠCong eqτ₁ eqτ₂)
equalTy? eqΓ (El t¹) (El t²) with equalTm? (✶Cong eqΓ) t¹ t² =>
  _ _   _       _       notEqualTm    = notEqualTy
  _ eqΓ (El t¹) (El t²) (equalTm eqt) = equalTy (ElCong eqt)
equalTy? _ _ _ = notEqualTy

(≟) (Γ : Ctxt) : (τ¹, τ² : Ty Γ) -> TyEq? τ¹ τ²
(≟) {Γ} = equalTy? (reflCtxt Γ)

(≜?) : (Γ¹, Γ² : Ctxt) -> CtxtEq? Γ¹ Γ²
ε         ≜? ε         = equalCtxt εCong
(Γ¹ ▻ σ¹) ≜? (Γ² ▻ σ²) with Γ¹ ≜? Γ² =>
  _ _         _         notEqualCtxt    = notEqualCtxt
  _ (Γ¹ ▻ σ¹) (Γ² ▻ σ²) (equalCtxt eqΓ) with equalTy? eqΓ σ¹ σ² =>
    _ _         _         _               notEqualTy    = notEqualCtxt
    _ (Γ¹ ▻ σ¹) (Γ² ▻ σ²) (equalCtxt eqΓ) (equalTy eqσ) =
      equalCtxt (eqΓ ▻Cong eqσ)
_ ≜? _ = notEqualCtxt

------------------------------------------------------------------------
-- Temporarily postulated congruence, required for completeness

evalCong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
         (tv¹ : TmVar) (t⁻¹ : Tm⁻ tv¹ t¹)
         (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
         (tv² : TmVar) (t⁻² : Tm⁻ tv² t²)
  : t⁻¹ ≈⊢⁻ t⁻² -> eval t⁻¹ ≈⊨ eval t⁻²

------------------------------------------------------------------------
-- Types expressing that a possible equality actually is an equality

data VarEq! {Γ¹ : Ctxt} {τ¹ : Ty Γ¹}
            {Γ² : Ctxt} {τ² : Ty Γ²}
            {v¹ : Γ¹ ∋ τ¹} {v² : Γ² ∋ τ²}
            (eqv? : VarEq? v¹ v²) : Set where
  varEq! : (eqv : v¹ ≈∋ v²) -> eqv? ≡ equalVar eqv -> VarEq! eqv?

data NFEq! {Γ¹ : Ctxt} {τ¹ : Ty Γ¹} {t¹ : Γ¹ ⊢ τ¹}
           {Γ² : Ctxt} {τ² : Ty Γ²} {t² : Γ² ⊢ τ²}
           {nf¹ : NF t¹} {nf² : NF t²}
           (eqnf? : NFEq? nf¹ nf²) : Set where
  nfEq! : (eqnf : nf¹ ≈⊨ nf²) -> eqnf? ≡ equalNF eqnf -> NFEq! eqnf?

data AtomEq! {Γ¹ : Ctxt} {τ¹ : Ty Γ¹} {t¹ : Γ¹ ⊢ τ¹}
             {Γ² : Ctxt} {τ² : Ty Γ²} {t² : Γ² ⊢ τ²}
             {at¹ : Atom t¹} {at² : Atom t²}
             (eqat? : AtomEq? at¹ at²) : Set where
  atomEq! :  (eqat : at¹ ≈⊨′ at²) -> eqat? ≡ equalAtom eqat
          -> AtomEq! eqat?

data Tm⁻Eq! {Γ¹ : Ctxt} {τ¹ : Ty Γ¹} {t¹ : Γ¹ ⊢ τ¹} {tv¹ : TmVar}
            {Γ² : Ctxt} {τ² : Ty Γ²} {t² : Γ² ⊢ τ²} {tv² : TmVar}
            {t⁻¹ : Tm⁻ tv¹ t¹} {t⁻² : Tm⁻ tv² t²}
            (eqt⁻? : Tm⁻Eq? t⁻¹ t⁻²) : Set where
  tm⁻Eq! : (eqt⁻ : t⁻¹ ≈⊢⁻ t⁻²) -> eqt⁻? ≡ equalTm⁻ eqt⁻ -> Tm⁻Eq! eqt⁻?

data TmEq! {Γ¹ : Ctxt} {τ¹ : Ty Γ¹}
           {Γ² : Ctxt} {τ² : Ty Γ²}
           {t¹ : Γ¹ ⊢ τ¹} {t² : Γ² ⊢ τ²}
           (eqt? : TmEq? t¹ t²) : Set where
  tmEq! : (eqt : t¹ ≈⊢ t²) -> eqt? ≡ equalTm eqt -> TmEq! eqt?

data TyEq! {Γ¹, Γ² : Ctxt} {τ¹ : Ty Γ¹} {τ² : Ty Γ²}
           (eqτ? : TyEq? τ¹ τ²) : Set where
  tyEq! : (eqτ : τ¹ ≛ τ²) -> eqτ? ≡ equalTy eqτ -> TyEq! eqτ?

data CtxtEq! {Γ¹, Γ² : Ctxt} (eqΓ? : CtxtEq? Γ¹ Γ²) : Set where
  ctxtEq! : (eqΓ : Γ¹ ≜ Γ²) -> eqΓ? ≡ equalCtxt eqΓ -> CtxtEq! eqΓ?

------------------------------------------------------------------------
-- Completeness

abstract

  varEqComplete (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
                (Γ² : Ctxt) (τ² : Ty Γ²)
    :  (eqΓ : Γ¹ ≜ Γ²) -> (v¹ : Γ¹ ∋ τ¹) -> (v² : Γ² ∋ τ²) -> v¹ ≈∋ v²
    -> VarEq! (equalVar? eqΓ v¹ v²)
  varEqComplete (eqΓ ▻Cong eqσ) (vz σ¹)    (vz σ²)    (vzCong eqσ') =
    varEq! (vzCong eqσ) refl
  varEqComplete (eqΓ ▻Cong eqσ) (vs v¹ σ¹) (vs v² σ²) (vsCong eqv' eqσ')
    with varEqComplete eqΓ v¹ v² eqv' =>
    _ (eqΓ ▻Cong eqσ) (vs v¹ σ¹) (vs v² σ²) (vsCong eqv' eqσ')
      (varEq! eqv eqv?) =
        varEq! (vsCong eqv eqσ) rewrite-plugin(eqv?, refl)
  varEqComplete eqΓ (v¹ ∷≡∋ eqτ) v² (castVarEqˡ eqτ eqv')
    with varEqComplete eqΓ v¹ v² eqv' =>
    _ eqΓ (v¹ ∷≡∋ eqτ) v² (castVarEqˡ eqτ eqv') (varEq! eqv eqv?) =
      varEq! (castVarEqˡ _ eqv) rewrite-plugin(eqv?, refl)
  varEqComplete eqΓ (vz σ¹) (v² ∷≡∋ eqτ₂) (castVarEqʳ eqτ₂ eqv')
    with varEqComplete eqΓ (vz σ¹) v² eqv' =>
    _ eqΓ (vz σ¹) (v² ∷≡∋ eqτ₂) (castVarEqʳ eqτ₂ eqv')
      (varEq! eqv eqv?) =
      varEq! (castVarEqʳ eqτ₂ eqv) rewrite-plugin(eqv?, refl)
  varEqComplete eqΓ (vs v¹ σ¹) (v² ∷≡∋ eqτ₂) (castVarEqʳ eqτ₂ eqv')
    with varEqComplete eqΓ (vs v¹ σ¹) v² eqv' =>
    _ eqΓ (vs v¹ σ¹) (v² ∷≡∋ eqτ₂) (castVarEqʳ eqτ₂ eqv')
      (varEq! eqv eqv?) =
      varEq! (castVarEqʳ eqτ₂ eqv) rewrite-plugin(eqv?, refl)
  varEqComplete eqΓ (v¹ ∷≡∋ eqτ₁) (v² ∷≡∋ eqτ₂) (castVarEqʳ eqτ₂ eqv')
    with varEqComplete eqΓ v¹ (v² ∷≡∋ eqτ₂)
                       (symVar (castVarEq v¹ eqτ₁)
                          `transVar`
                        castVarEqʳ eqτ₂ eqv') =>
    _ eqΓ (v¹ ∷≡∋ eqτ₁) (v² ∷≡∋ eqτ₂) (castVarEqʳ eqτ₂ eqv')
      (varEq! eqv eqv?) =
      varEq! (castVarEqˡ eqτ₁ eqv) rewrite-plugin(eqv?, refl)

mutual

  nfEqComplete (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    :  (eqτ : τ¹ ≛ τ²) -> (nf¹ : NF t¹) -> (nf² : NF t²)
    -> nf¹ ≈⊨ nf² -> NFEq! (equalNF? eqτ nf¹ nf²)
  nfEqComplete = helper
    where
    abstract
      helper (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
        :  (eqτ : τ¹ ≛ τ²) -> (nf¹ : NF t¹) -> (nf² : NF t²)
        -> nf¹ ≈⊨ nf² -> NFEq! (equalNF? eqτ nf¹ nf²)
      helper eqτ (✶Atom at¹) (✶Atom at²) (✶AtomCong eqat')
        with atomEqComplete (tyEqCtxtEq eqτ) at¹ at² eqat' =>
        _ eqτ (✶Atom at¹) (✶Atom at²) (✶AtomCong eqat')
          (atomEq! eqat eqat?) =
          nfEq! (✶AtomCong eqat) rewrite-plugin(eqat?, refl)
      helper eqτ (ElAtom at¹) (ElAtom at²) (ElAtomCong eqat')
        with atomEqComplete (tyEqCtxtEq eqτ) at¹ at² eqat' =>
        _ eqτ (ElAtom at¹) (ElAtom at²) (ElAtomCong eqat')
          (atomEq! eqat eqat?) =
          nfEq! (ElAtomCong eqat) rewrite-plugin(eqat?, refl)
      helper (ΠCong eqτ₁ eqτ₂) (λ⊨ nf¹) (λ⊨ nf²) (λ⊨Cong eqnf')
        with nfEqComplete eqτ₂ nf¹ nf² eqnf' =>
        _ (ΠCong eqτ₁ eqτ₂) (λ⊨ nf¹) (λ⊨ nf²) (λ⊨Cong eqnf')
          (nfEq! eqnf eqnf?) =
          nfEq! (λ⊨Cong eqnf) rewrite-plugin(eqnf?, refl)
      helper eqτ (nf¹ ∷≡⊨ eqt) nf² (castNFEqˡ eqt eqnf')
        with nfEqComplete (tmEqTyEq eqt `transTy` eqτ)
                          nf¹ nf² eqnf' =>
        _ eqτ (nf¹ ∷≡⊨ eqt) nf² (castNFEqˡ eqt eqnf')
          (nfEq! eqnf eqnf?) =
          nfEq! (castNFEqˡ eqt eqnf) rewrite-plugin(eqnf?, refl)
      helper eqτ (✶Atom at¹) (nf² ∷≡⊨ eqt) (castNFEqʳ eqt eqnf')
        with nfEqComplete (eqτ `transTy` symTy (tmEqTyEq eqt))
                          (✶Atom at¹) nf² eqnf' =>
        _ eqτ (✶Atom at¹) (nf² ∷≡⊨ eqt) (castNFEqʳ eqt eqnf')
          (nfEq! eqnf eqnf?) =
          nfEq! (castNFEqʳ eqt eqnf) rewrite-plugin(eqnf?, refl)
      helper eqτ (ElAtom at¹) (nf² ∷≡⊨ eqt) (castNFEqʳ eqt eqnf')
        with nfEqComplete (eqτ `transTy` symTy (tmEqTyEq eqt))
                          (ElAtom at¹) nf² eqnf' =>
        _ eqτ (ElAtom at¹) (nf² ∷≡⊨ eqt) (castNFEqʳ eqt eqnf')
          (nfEq! eqnf eqnf?) =
          nfEq! (castNFEqʳ eqt eqnf) rewrite-plugin(eqnf?, refl)
      helper eqτ (λ⊨ nf¹) (nf² ∷≡⊨ eqt) (castNFEqʳ eqt eqnf')
        with nfEqComplete (eqτ `transTy` symTy (tmEqTyEq eqt))
                          (λ⊨ nf¹) nf² eqnf' =>
        _ eqτ (λ⊨ nf¹) (nf² ∷≡⊨ eqt) (castNFEqʳ eqt eqnf')
          (nfEq! eqnf eqnf?) =
          nfEq! (castNFEqʳ eqt eqnf) rewrite-plugin(eqnf?, refl)
      helper eqτ (nf¹ ∷≡⊨ eqt₁) (nf² ∷≡⊨ eqt₂) (castNFEqʳ eqt₂ eqnf')
        with nfEqComplete (tmEqTyEq eqt₁ `transTy` eqτ)
                          nf¹ (nf² ∷≡⊨ eqt₂)
                          (symNF (castNFEq nf¹ eqt₁)
                            `transNF`
                          castNFEqʳ eqt₂ eqnf') =>
        _ eqτ (nf¹ ∷≡⊨ eqt₁) (nf² ∷≡⊨ eqt₂) (castNFEqʳ eqt₂ eqnf')
          (nfEq! eqnf eqnf?) =
          nfEq! (castNFEqˡ eqt₁ eqnf) rewrite-plugin(eqnf?, refl)

  atomEqComplete (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
                 (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    :  (eqΓ : Γ¹ ≜ Γ²) -> (at¹ : Atom t¹) -> (at² : Atom t²)
    -> at¹ ≈⊨′ at² -> AtomEq! (equalAtom? eqΓ at¹ at²)
  atomEqComplete = helper
    where
    abstract
      helper (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
        :  (eqΓ : Γ¹ ≜ Γ²) -> (at¹ : Atom t¹) -> (at² : Atom t²)
        -> at¹ ≈⊨′ at² -> AtomEq! (equalAtom? eqΓ at¹ at²)
      helper eqΓ (var⊨ v¹) (var⊨ v²) (var⊨Cong eqv')
        with varEqComplete eqΓ v¹ v² eqv' =>
        _ eqΓ (var⊨ v¹) (var⊨ v²) (var⊨Cong eqv') (varEq! eqv eqv?) =
          atomEq! (var⊨Cong eqv) rewrite-plugin(eqv?, refl)
      helper eqΓ (at₁¹ @⊨ nf₂¹) (at₁² @⊨ nf₂²)
             (app⊨Cong eqat₁' eqnf₂')
        with atomEqComplete eqΓ at₁¹ at₁² eqat₁' =>
        _ eqΓ (at₁¹ @⊨ nf₂¹) (at₁² @⊨ nf₂²) (app⊨Cong eqat₁' eqnf₂')
          (atomEq! eqat₁ eqat₁?)
          with nfEqComplete (ΠEqLeftEq $ tmEqTyEq (atomEqTmEq eqat₁))
               nf₂¹ nf₂² eqnf₂' =>
          _ eqΓ (at₁¹ @⊨ nf₂¹) (at₁² @⊨ nf₂²) (app⊨Cong eqat₁' eqnf₂')
            (atomEq! eqat₁ eqat₁?) (nfEq! eqnf₂ eqnf₂?) =
            atomEq! (app⊨Cong eqat₁ eqnf₂)
                    rewrite-plugin(eqat₁?, rewrite-plugin(eqnf₂?, refl))
      helper eqΓ (at¹ ∷≡⊨′ eqt) at² (castAtomEqˡ eqt eqat')
        with atomEqComplete eqΓ at¹ at² eqat' =>
        _ eqΓ (at¹ ∷≡⊨′ eqt) at² (castAtomEqˡ eqt eqat')
          (atomEq! eqat eqat?) =
          atomEq! (castAtomEqˡ eqt eqat) rewrite-plugin(eqat?, refl)
      helper eqΓ (var⊨ v¹) (at² ∷≡⊨′ eqt) (castAtomEqʳ eqt eqat')
        with atomEqComplete eqΓ (var⊨ v¹) at² eqat' =>
        _ eqΓ (var⊨ v¹) (at² ∷≡⊨′ eqt) (castAtomEqʳ eqt eqat')
          (atomEq! eqat eqat?) =
          atomEq! (castAtomEqʳ eqt eqat) rewrite-plugin(eqat?, refl)
      helper eqΓ (at₁¹ @⊨ nf₂¹) (at² ∷≡⊨′ eqt)
             (castAtomEqʳ eqt eqat')
        with atomEqComplete eqΓ (at₁¹ @⊨ nf₂¹) at² eqat' =>
        _ eqΓ (at₁¹ @⊨ nf₂¹) (at² ∷≡⊨′ eqt) (castAtomEqʳ eqt eqat')
          (atomEq! eqat eqat?) =
          atomEq! (castAtomEqʳ eqt eqat) rewrite-plugin(eqat?, refl)
      helper eqΓ (at¹ ∷≡⊨′ eqt₁) (at² ∷≡⊨′ eqt₂)
             (castAtomEqʳ eqt₂ eqat')
        with atomEqComplete eqΓ at¹ (at² ∷≡⊨′ eqt₂)
                            (symAtom (castAtomEq at¹ eqt₁)
                               `transAtom`
                             castAtomEqʳ eqt₂ eqat') =>
        _ eqΓ (at¹ ∷≡⊨′ eqt₁) (at² ∷≡⊨′ eqt₂)
          (castAtomEqʳ eqt₂ eqat') (atomEq! eqat eqat?) =
          atomEq! (castAtomEqˡ eqt₁ eqat) rewrite-plugin(eqat?, refl)

abstract

  tm⁻EqComplete (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
                (tv¹ : TmVar) (t⁻¹ : Tm⁻ tv¹ t¹)
                (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
                (tv² : TmVar) (t⁻² : Tm⁻ tv² t²)
    : (eqτ : τ¹ ≛ τ²) -> t⁻¹ ≈⊢⁻ t⁻² -> Tm⁻Eq! (equalTm⁻? eqτ t⁻¹ t⁻²)
  tm⁻EqComplete eqτ eqt⁻ with nfEqComplete eqτ _ _ (evalCong eqt⁻) =>
    _ {Γ¹} {τ¹} {t¹} {tv¹} {t⁻¹} {Γ²} {τ²} {t²} {tv²} {t⁻²}
      eqτ eqt⁻ (nfEq! eqnf eqnf?) =
      tm⁻Eq! (tmEqTm⁻Eq $ nfEqTmEq eqnf)
             -- For some reason rewrite-plugin didn't work here.
             (subst (\eqnf? ->
                        equalTm⁻?₁ Γ¹ τ¹ t¹ tv¹ Γ² τ² t² tv²
                                   eqτ t⁻¹ t⁻² eqnf? ≡
                        equalTm⁻ (tmEqTm⁻Eq $ nfEqTmEq eqnf))
                    (sym eqnf?) refl)

  tmEqComplete (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : (eqτ : τ¹ ≛ τ²) -> t¹ ≈⊢ t² -> TmEq! (equalTm? eqτ t¹ t²)
  tmEqComplete eqτ eqt with tm⁻EqComplete eqτ (tmToTm⁻Cong eqt) =>
    _ eqτ eqt (tm⁻Eq! eqt⁻ eqt⁻?) =
      tmEq! (tm⁻EqTmEq eqt⁻) rewrite-plugin(eqt⁻?, refl)

  tyEqComplete' (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    : (eqΓ : Γ¹ ≜ Γ²) -> τ¹ ≛ τ² -> TyEq! (equalTy? eqΓ τ¹ τ²)
  tyEqComplete' eqΓ (✶Cong eqΓ')        = tyEq! (✶Cong eqΓ) refl
  tyEqComplete' eqΓ (ΠCong eqτ₁' eqτ₂') with tyEqComplete' eqΓ eqτ₁' =>
    _ eqΓ (ΠCong _ eqτ₂') (tyEq! eqτ₁ eqτ₁?)
      with tyEqComplete' (eqΓ ▻Cong eqτ₁) eqτ₂' =>
      _ eqΓ (ΠCong _ _) (tyEq! eqτ₁ eqτ₁?) (tyEq! eqτ₂ eqτ₂?) =
        tyEq! (ΠCong eqτ₁ eqτ₂)
              rewrite-plugin(eqτ₁?, rewrite-plugin(eqτ₂?, refl))
  tyEqComplete' eqΓ (ElCong eqt') with tmEqComplete (✶Cong eqΓ) eqt' =>
    _ eqΓ (ElCong _) (tmEq! eqt eqt?) =
      tyEq! (ElCong eqt) rewrite-plugin(eqt?, refl)

  tyEqComplete (Γ : Ctxt) (τ¹, τ² : Ty Γ)
    : τ¹ ≛ τ² -> TyEq! (τ¹ ≟ τ²)
  tyEqComplete {Γ} eqτ = tyEqComplete' (reflCtxt Γ) eqτ

  ctxtEqComplete (Γ¹, Γ² : Ctxt)
    : Γ¹ ≜ Γ² -> CtxtEq! (Γ¹ ≜? Γ²)
  ctxtEqComplete εCong             = ctxtEq! εCong refl
  ctxtEqComplete (eqΓ' ▻Cong eqσ') with ctxtEqComplete eqΓ' =>
    _ (_ ▻Cong eqσ') (ctxtEq! eqΓ eqΓ?) with tyEqComplete' eqΓ eqσ' =>
      _ (_ ▻Cong _) (ctxtEq! eqΓ eqΓ?) (tyEq! eqσ eqσ?) =
        ctxtEq! (eqΓ ▻Cong eqσ)
                rewrite-plugin(eqΓ?, rewrite-plugin(eqσ?, refl))

DecidableEquality : Type
DecidableEquality = time-plugin(DecidableEquality, Set)
