------------------------------------------------------------------------
-- Propositional equality
------------------------------------------------------------------------

-- Based on a propositional equality definition in the agdaLight CVS
-- repository.

infix 40 ≡, ==

------------------------------------------------------------------------
-- Propositional equality

data (==) {A : Set} (a : A) : A -> Set where
  refl : a == a

(≡) (A : Set) : A -> A -> Set
(≡) = (==)

------------------------------------------------------------------------
-- The equality is an equivalence

sym (a : Set) (x, y : a) : x ≡ y -> y ≡ x
sym refl = refl

trans (a : Set) (x, y, z : a) : x ≡ y -> y ≡ z -> x ≡ z
trans refl refl = refl

------------------------------------------------------------------------
-- The equality is substitutive

subst (a : Set) (x, y : a) : (P : a -> Set) -> x ≡ y -> P x -> P y
subst _ refl p = p

substBin (a, b : Set) (x₁, y₁ : a) (x₂, y₂ : b)
  : (P : a -> b -> Set) -> x₁ ≡ y₁ -> x₂ ≡ y₂ -> P x₁ x₂ -> P y₁ y₂
substBin _ refl refl p = p

------------------------------------------------------------------------
-- The equality is a congruence

cong (a, b : Set) (x, y : a) : (f : a -> b) -> x ≡ y -> f x ≡ f y
cong _ refl = refl

congBin (a, b, c : Set) (x₁, y₁ : a) (x₂, y₂ : b)
  : ((+) : a -> b -> c) -> x₁ ≡ y₁ -> x₂ ≡ y₂ -> (x₁ + x₂) ≡ (y₁ + y₂)
congBin _ refl refl = refl

------------------------------------------------------------------------
-- Miscellaneous lemmas and helpers

removeSubst (a : Set) (x, y : a)
  :  (P : a -> Set) -> (f : (z : a) -> P z) -> (eq : x ≡ y)
  -> subst P eq (f x) ≡ f y
removeSubst _ _ refl = refl

toPropEq (a : Set) (x, y : a) : x ≡ y -> x == y
toPropEq refl = refl

data Inspect {a : Set} (x : a) : Set where
  (:≡) : (y : a) -> x ≡ y -> Inspect x

inspect (a : Set) : (x : a) -> Inspect x
inspect x = x :≡ refl
