------------------------------------------------------------------------
-- The various equalities are equivalence relations
------------------------------------------------------------------------

-- # import Explicit.Types

infixl 47 `transCtxt`, `transRCtxt`, `transTy`, `transSubst`,
          `transVar`

------------------------------------------------------------------------
-- A congruence

abstract

  castVarCong (Γ¹, Γ² : Ctxt) (τ¹¹, τ¹² : Ty Γ¹) (τ²¹, τ²² : Ty Γ²)
              (v¹ : Γ¹ ∋ τ¹¹) (v² : Γ² ∋ τ²¹)
    :  (eqτ¹ : τ¹¹ ≛ τ¹²) -> (eqτ² : τ²¹ ≛ τ²²)
    -> v¹ ≈∋ v² -> v¹ ∷≡∋ eqτ¹ ≈∋ v² ∷≡∋ eqτ²
  castVarCong eqτ¹ eqτ² eqv = castVarEqʳ eqτ² $ castVarEqˡ eqτ¹ eqv

------------------------------------------------------------------------
-- Symmetry

mutual

  symCtxt (Γ¹, Γ² : Ctxt) : Γ¹ ≜ Γ² -> Γ² ≜ Γ¹
  symCtxt εCong           = εCong
  symCtxt (eqΓ ▻Cong eqτ) = symCtxt eqΓ ▻Cong symTy eqτ

  symTy (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
        (Γ² : Ctxt) (τ² : Ty Γ²)
    : τ¹ ≛ τ² -> τ² ≛ τ¹
  symTy (✶Cong eqΓ)       = ✶Cong (symCtxt eqΓ)
  symTy (ΠCong eqτ₁ eqτ₂) = ΠCong (symTy eqτ₁) (symTy eqτ₂)
  symTy (ElCong eqt)      = ElCong (symTm eqt)

abstract

  symRCtxt (Γ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹)
           (Γ² : Ctxt) (Γ'² : RCtxt Γ²)
    : Γ'¹ ≜→ Γ'² -> Γ'² ≜→ Γ'¹
  symRCtxt (εʳCong eqΓ)      = εʳCong (symCtxt eqΓ)
  symRCtxt (eqΓ' ▻ʳCong eqτ) = symRCtxt eqΓ' ▻ʳCong symTy eqτ

  symSubst (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
           (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
    : ρ¹ ≘ ρ² -> ρ² ≘ ρ¹
  symSubst (emptyCong eqΔ)        = emptyCong (symCtxt eqΔ)
  symSubst (snocCong eqρ eqτ eqt) =
    snocCong (symSubst eqρ) (symTy eqτ) (symTm eqt)

  symVar (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
         (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
    : v¹ ≈∋ v² -> v² ≈∋ v¹
  symVar (vzCong eqσ)         = vzCong (symTy eqσ)
  symVar (vsCong eqv eqσ)     = vsCong (symVar eqv) (symTy eqσ)
  symVar (castVarEqˡ eqτ eqv) = castVarEqʳ eqτ (symVar eqv)
  symVar (castVarEqʳ eqτ eqv) = castVarEqˡ eqτ (symVar eqv)

------------------------------------------------------------------------
-- Transitivity

mutual

  transCtxt (Γ¹, Γ², Γ³ : Ctxt) : Γ¹ ≜ Γ² -> Γ² ≜ Γ³ -> Γ¹ ≜ Γ³
  transCtxt εCong               εCong               = εCong
  transCtxt (eqΓ¹² ▻Cong eqτ¹²) (eqΓ²³ ▻Cong eqτ²³) =
    transCtxt eqΓ¹² eqΓ²³ ▻Cong transTy eqτ¹² eqτ²³

  transTy (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
          (Γ² : Ctxt) (τ² : Ty Γ²)
          (Γ³ : Ctxt) (τ³ : Ty Γ³)
    : τ¹ ≛ τ² -> τ² ≛ τ³ -> τ¹ ≛ τ³
  transTy (✶Cong eqΓ¹²) (✶Cong eqΓ²³) = ✶Cong (transCtxt eqΓ¹² eqΓ²³)
  transTy (ΠCong eqτ₁¹² eqτ₂¹²) (ΠCong eqτ₁²³ eqτ₂²³) =
    ΠCong (transTy eqτ₁¹² eqτ₁²³) (transTy eqτ₂¹² eqτ₂²³)
  transTy (ElCong eqt¹²) (ElCong eqt²³) = ElCong (transTm eqt¹² eqt²³)

abstract

  transRCtxt (Γ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹)
             (Γ² : Ctxt) (Γ'² : RCtxt Γ²)
             (Γ³ : Ctxt) (Γ'³ : RCtxt Γ³)
    : Γ'¹ ≜→ Γ'² -> Γ'² ≜→ Γ'³ -> Γ'¹ ≜→ Γ'³
  transRCtxt (εʳCong eqΓ¹²) (εʳCong eqΓ²³) =
   εʳCong (transCtxt eqΓ¹² eqΓ²³)
  transRCtxt (eqΓ'¹² ▻ʳCong eqτ¹²) (eqΓ'²³ ▻ʳCong eqτ²³) =
    transRCtxt eqΓ'¹² eqΓ'²³ ▻ʳCong transTy eqτ¹² eqτ²³

  transSubst (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
             (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
             (Γ³, Δ³ : Ctxt) (ρ³ : Subst Γ³ Δ³)
    : ρ¹ ≘ ρ² -> ρ² ≘ ρ³ -> ρ¹ ≘ ρ³
  transSubst (emptyCong eqΔ¹²) (emptyCong eqΔ²³) =
    emptyCong (transCtxt eqΔ¹² eqΔ²³)
  transSubst (snocCong eqρ¹² eqτ¹² eqt¹²) (snocCong eqρ²³ eqτ²³ eqt²³) =
    snocCong (transSubst eqρ¹² eqρ²³)
             (transTy eqτ¹² eqτ²³)
             (transTm eqt¹² eqt²³)

  transVar (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
           (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
           (Γ³ : Ctxt) (τ³ : Ty Γ³) (v³ : Γ³ ∋ τ³)
    : v¹ ≈∋ v² -> v² ≈∋ v³ -> v¹ ≈∋ v³
  transVar (vzCong eqσ¹²) (vzCong eqσ²³) = vzCong (transTy eqσ¹² eqσ²³)
  transVar (vsCong eqv¹² eqσ¹²) (vsCong eqv²³ eqσ²³) =
    vsCong (transVar eqv¹² eqv²³) (transTy eqσ¹² eqσ²³)
  transVar (castVarEqˡ eqτ eqv¹) eqv² =
    castVarEqˡ eqτ (transVar eqv¹ eqv²)
  transVar eqv¹ (castVarEqʳ eqτ eqv²) =
    castVarEqʳ eqτ (transVar eqv¹ eqv²)
  transVar (castVarEqʳ eqτ eqv₁) (castVarEqˡ eqτ eqv₂) =
    transVar eqv₁ eqv₂

------------------------------------------------------------------------
-- Reflexivity

mutual

  reflCtxt : (Γ : Ctxt) -> Γ ≜ Γ
  reflCtxt ε       = εCong
  reflCtxt (Γ ▻ τ) = reflCtxt Γ ▻Cong reflTy τ

  reflTy (Γ : Ctxt) : (τ : Ty Γ) -> τ ≛ τ
  reflTy (✶ Γ)     = ✶Cong (reflCtxt Γ)
  reflTy (Π τ₁ τ₂) = ΠCong (reflTy τ₁) (reflTy τ₂)
  reflTy (El t)    = ElCong (reflTm t)

abstract

  reflRCtxt (Γ : Ctxt) : (Γ' : RCtxt Γ) -> Γ' ≜→ Γ'
  reflRCtxt (εʳ Γ)    = εʳCong (reflCtxt Γ)
  reflRCtxt (Γ' ▻ʳ τ) = reflRCtxt Γ' ▻ʳCong reflTy τ

  reflVar (Γ : Ctxt) (τ : Ty Γ)
    : (v : Γ ∋ τ) -> v ≈∋ v
  reflVar (vz σ)      = vzCong (reflTy σ)
  reflVar (vs v σ)    = vsCong (reflVar v) (reflTy σ)
  reflVar (v ∷≡∋ eqτ) = castVarCong eqτ eqτ (reflVar v)

  reflSubst (Γ, Δ : Ctxt)
    : (ρ : Subst Γ Δ) -> ρ ≘ ρ
  reflSubst {ε}     {Δ} ρ = emptyCong (reflCtxt Δ)
  reflSubst {Γ ▻ τ} {Δ} ρ =
    snocCong (reflSubst {Γ} {Δ} $ init ρ) (reflTy τ) (reflTm $ last' ρ)

------------------------------------------------------------------------
-- Some "between" lemmas

abstract

  betweenCtxt (Γ₁, Γ₂, Γ₃, Γ₄ : Ctxt)
    : Γ₁ ≜ Γ₂ -> Γ₄ ≜ Γ₃ -> Γ₂ ≜ Γ₃ -> Γ₁ ≜ Γ₄
  betweenCtxt eq₁ eq₃ eq₂ = eq₁ `transCtxt` eq₂ `transCtxt` symCtxt eq₃

  betweenRCtxt (Γ₁ : Ctxt) (Γ'₁ : RCtxt Γ₁)
               (Γ₂ : Ctxt) (Γ'₂ : RCtxt Γ₂)
               (Γ₃ : Ctxt) (Γ'₃ : RCtxt Γ₃)
               (Γ₄ : Ctxt) (Γ'₄ : RCtxt Γ₄)
    : Γ'₁ ≜→ Γ'₂ -> Γ'₄ ≜→ Γ'₃ -> Γ'₂ ≜→ Γ'₃ -> Γ'₁ ≜→ Γ'₄
  betweenRCtxt eq₁ eq₃ eq₂ =
    eq₁ `transRCtxt` eq₂ `transRCtxt` symRCtxt eq₃

  betweenTy (Γ₁ : Ctxt) (τ₁ : Ty Γ₁)
            (Γ₂ : Ctxt) (τ₂ : Ty Γ₂)
            (Γ₃ : Ctxt) (τ₃ : Ty Γ₃)
            (Γ₄ : Ctxt) (τ₄ : Ty Γ₄)
    : τ₁ ≛ τ₂ -> τ₄ ≛ τ₃ -> τ₂ ≛ τ₃ -> τ₁ ≛ τ₄
  betweenTy eq₁ eq₃ eq₂ = eq₁ `transTy` eq₂ `transTy` symTy eq₃

  betweenSubst (Γ₁, Δ₁ : Ctxt) (ρ₁ : Subst Γ₁ Δ₁)
               (Γ₂, Δ₂ : Ctxt) (ρ₂ : Subst Γ₂ Δ₂)
               (Γ₃, Δ₃ : Ctxt) (ρ₃ : Subst Γ₃ Δ₃)
               (Γ₄, Δ₄ : Ctxt) (ρ₄ : Subst Γ₄ Δ₄)
    : ρ₁ ≘ ρ₂ -> ρ₄ ≘ ρ₃ -> ρ₂ ≘ ρ₃ -> ρ₁ ≘ ρ₄
  betweenSubst eq₁ eq₃ eq₂ =
    eq₁ `transSubst` eq₂ `transSubst` symSubst eq₃

  betweenTm (Γ₁ : Ctxt) (τ₁ : Ty Γ₁) (t₁ : Γ₁ ⊢ τ₁)
            (Γ₂ : Ctxt) (τ₂ : Ty Γ₂) (t₂ : Γ₂ ⊢ τ₂)
            (Γ₃ : Ctxt) (τ₃ : Ty Γ₃) (t₃ : Γ₃ ⊢ τ₃)
            (Γ₄ : Ctxt) (τ₄ : Ty Γ₄) (t₄ : Γ₄ ⊢ τ₄)
    : t₁ ≈⊢ t₂ -> t₄ ≈⊢ t₃ -> t₂ ≈⊢ t₃ -> t₁ ≈⊢ t₄
  betweenTm eq₁ eq₃ eq₂ = eq₁ `transTm` eq₂ `transTm` symTm eq₃

------------------------------------------------------------------------
-- Some cast lemmas

abstract

  castVarEq (Γ : Ctxt) (τ¹, τ² : Ty Γ)
    :  (v : Γ ∋ τ¹) -> (eqτ : τ¹ ≛ τ²)
    -> v ∷≡∋ eqτ ≈∋ v
  castVarEq v eqτ = castVarEqˡ eqτ (reflVar v)

  substCastTm (Γ, Δ : Ctxt) (τ¹, τ² : Ty Γ)
    :  (t : Γ ⊢ τ¹) -> (eqτ : τ¹ ≛ τ²) -> (ρ : Subst Γ Δ)
    -> (t ∷≡⊢ eqτ) /⊢ ρ ≈⊢ t /⊢ ρ
  substCastTm t eqτ ρ = substTmCong (castTmEq t eqτ) (reflSubst ρ)

  substCastVar (Γ, Δ : Ctxt) (τ¹, τ² : Ty Γ)
    :  (v : Γ ∋ τ¹) -> (eqτ : τ¹ ≛ τ²) -> (ρ : Subst Γ Δ)
    -> var (v ∷≡∋ eqτ) /⊢ ρ ≈⊢ var v /⊢ ρ
  substCastVar v eqτ ρ =
    substTmCong (varCong $ castVarEq v eqτ) (reflSubst ρ)
