------------------------------------------------------------------------
-- Equal things have equal indices
------------------------------------------------------------------------

-- The lemmas below show that the heterogeneous equalities are
-- "morally homogeneous". For instance, if two terms are equal, then
-- their types are also equal.

-- # import Explicit.Types
-- # import Explicit.Equivalence
-- # import Explicit.Extensional
-- # import Explicit.SubstitutionLemmas

------------------------------------------------------------------------
-- Some equality projections

ΠEqLeftEq (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
          (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
  : Π τ₁¹ τ₂¹ ≛ Π τ₁² τ₂² -> τ₁¹ ≛ τ₁²
ΠEqLeftEq (ΠCong eqτ₁ eqτ₂) = eqτ₁

ΠEqRightEq (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
           (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
  : Π τ₁¹ τ₂¹ ≛ Π τ₁² τ₂² -> τ₂¹ ≛ τ₂²
ΠEqRightEq (ΠCong eqτ₁ eqτ₂) = eqτ₂

snocEqInitEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²)
  : Γ¹ ▻ τ¹ ≜ Γ² ▻ τ² -> Γ¹ ≜ Γ²
snocEqInitEq (eqΓ ▻Cong eqτ) = eqΓ

snocEqLastEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²)
  : Γ¹ ▻ τ¹ ≜ Γ² ▻ τ² -> τ¹ ≛ τ²
snocEqLastEq (eqΓ ▻Cong eqτ) = eqτ

------------------------------------------------------------------------
-- Only things with equal contexts can be equal

abstract

  rCtxtEqCtxtEq (Γ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹)
                (Γ² : Ctxt) (Γ'² : RCtxt Γ²)
    : Γ'¹ ≜→ Γ'² -> Γ¹ ≜ Γ²
  rCtxtEqCtxtEq (εʳCong eqΓ)      = eqΓ
  rCtxtEqCtxtEq (eqΓ' ▻ʳCong eqτ) = rCtxtEqCtxtEq eqΓ'

mutual

  tyEqCtxtEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²)
    : τ¹ ≛ τ² -> Γ¹ ≜ Γ²
  tyEqCtxtEq (✶Cong eqΓ)       = eqΓ
  tyEqCtxtEq (ΠCong eqτ₁ eqτ₂) = tyEqCtxtEq eqτ₁
  tyEqCtxtEq (ElCong eqt)      = tmEqCtxtEq eqt

  varEqCtxtEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
    : v¹ ≈∋ v² -> Γ¹ ≜ Γ²
  varEqCtxtEq (vzCong eqσ)         = tyEqCtxtEq eqσ ▻Cong eqσ
  varEqCtxtEq (vsCong eqv eqσ)     = varEqCtxtEq eqv ▻Cong eqσ
  varEqCtxtEq (castVarEqˡ eqτ eqv) = varEqCtxtEq eqv
  varEqCtxtEq (castVarEqʳ eqτ eqv) = varEqCtxtEq eqv

  tmEqCtxtEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : t¹ ≈⊢ t² -> Γ¹ ≜ Γ²
  tmEqCtxtEq (reflTm {Γ} t)        = reflCtxt Γ
  tmEqCtxtEq (symTm eqt)           = symCtxt (tmEqCtxtEq eqt)
  tmEqCtxtEq (transTm eqt¹² eqt²³) =
    transCtxt (tmEqCtxtEq eqt¹²) (tmEqCtxtEq eqt²³)
  tmEqCtxtEq (varCong eqv)               = varEqCtxtEq eqv
  tmEqCtxtEq (λCong eqt)                 = snocEqInitEq $ tmEqCtxtEq eqt
  tmEqCtxtEq (appCong eqt₁ eqt₂)         = tmEqCtxtEq eqt₁
  tmEqCtxtEq (substTmCong eqt eqρ)       = substEqCtxtEqRight eqρ
  tmEqCtxtEq (castTmEq {Γ} t eqτ)        = reflCtxt Γ
  tmEqCtxtEq (β {Γ} t₁ t₂)               = reflCtxt Γ
  tmEqCtxtEq (η {Γ} t)                   = reflCtxt Γ
  tmEqCtxtEq (idVanishesTm {Γ} t)        = reflCtxt Γ
  tmEqCtxtEq (substWk {Γ} v σ)           = reflCtxt (Γ ▻ σ)
  tmEqCtxtEq (substVzSub {Γ} σ t)        = reflCtxt Γ
  tmEqCtxtEq (substVsSub {Γ} v σ t)      = reflCtxt Γ
  tmEqCtxtEq (substVzLift {Γ} {Δ} σ ρ)   = reflCtxt (Δ ▻ σ / ρ)
  tmEqCtxtEq (substVsLift {Γ} {Δ} v σ ρ) = reflCtxt (Δ ▻ σ / ρ)
  tmEqCtxtEq (substλ {Γ} {Δ} t ρ)        = reflCtxt Δ
  tmEqCtxtEq (substApp {Γ} {Δ} t₁ t₂ ρ)  = reflCtxt Δ
  tmEqCtxtEq (compSplitsTm {Γ} {Δ} {X} t ρ₁ ρ₂) = reflCtxt X

------------------------------------------------------------------------
-- A congruence used below

abstract

  wkCong (Γ¹ : Ctxt) (σ¹ : Ty Γ¹)
         (Γ² : Ctxt) (σ² : Ty Γ²)
    : σ¹ ≛ σ² -> wk σ¹ ≘ wk σ²
  wkCong eqσ = wkCong' (tyEqCtxtEq eqσ) eqσ

------------------------------------------------------------------------
-- Only things with equal types can be equal

abstract

  varEqTyEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
    : v¹ ≈∋ v² -> τ¹ ≛ τ²
  varEqTyEq (vzCong eqσ)     = substTyCong eqσ (wkCong eqσ)
  varEqTyEq (vsCong eqv eqσ) = substTyCong (varEqTyEq eqv) (wkCong eqσ)
  varEqTyEq (castVarEqˡ eqτ eqv) = symTy eqτ `transTy` varEqTyEq eqv
  varEqTyEq (castVarEqʳ eqτ eqv) = varEqTyEq eqv `transTy` eqτ

mutual

  tmEqTyEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
           (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : t¹ ≈⊢ t² -> τ¹ ≛ τ²
  tmEqTyEq (reflTm {Γ} {τ} t)    = reflTy τ
  tmEqTyEq (symTm eqt)           = symTy (tmEqTyEq eqt)
  tmEqTyEq (transTm eqt¹² eqt²³) =
    transTy (tmEqTyEq eqt¹²) (tmEqTyEq eqt²³)
  tmEqTyEq (varCong eqv) = varEqTyEq eqv
  tmEqTyEq (λCong eqt)   =
    ΠCong (snocEqLastEq $ tmEqCtxtEq eqt) (tmEqTyEq eqt)
  tmEqTyEq (appCong eqt₁ eqt₂) =
    substTyCong (ΠEqRightEq $ tmEqTyEq eqt₁) (subCong eqt₂)
  tmEqTyEq (substTmCong eqt eqρ)   = substTyCong (tmEqTyEq eqt) eqρ
  tmEqTyEq (castTmEq t eqτ)        = symTy eqτ
  tmEqTyEq (β {Γ} {τ₁} {τ₂} t₁ t₂) = reflTy (τ₂ / sub t₂)
  tmEqTyEq (η {Γ} {τ₁} {τ₂} t) = ΠCong (reflTy τ₁) (subVzWithVzTy τ₂)
  tmEqTyEq (idVanishesTm {Γ} {τ} t)             = idVanishesTy τ
  tmEqTyEq (substWk {Γ} {τ} v σ)                = reflTy (τ / wk σ)
  tmEqTyEq (substVzSub σ t)                     = wkSubVanishesTy σ t
  tmEqTyEq (substVsSub {Γ} {τ} v σ t)           = wkSubVanishesTy τ t
  tmEqTyEq (substVzLift σ ρ)                    = wkCommutesTy σ σ ρ
  tmEqTyEq (substVsLift {Γ} {Δ} {τ} v σ ρ)      = wkCommutesTy τ σ ρ
  tmEqTyEq (substλ {Γ} {Δ} {τ₁} {τ₂} t ρ)       = reflTy (Π τ₁ τ₂ / ρ)
  tmEqTyEq (substApp {Γ} {Δ} {τ₁} {τ₂} t₁ t₂ ρ) = subCommutesTy τ₂ t₂ ρ
  tmEqTyEq (compSplitsTm {Γ} {Δ} {X} {τ} t ρ₁ ρ₂) = compSplitsTy τ ρ₁ ρ₂

  -- Another congruence.

  subCong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
          (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : t¹ ≈⊢ t² -> sub t¹ ≘ sub t²
  subCong eqt = subCong' (tmEqCtxtEq eqt) (tmEqTyEq eqt) eqt
