------------------------------------------------------------------------
-- Soundness
------------------------------------------------------------------------

-- These lemmas show that this view is sound with respect to the
-- explicit substitution theory.

-- # import Explicit
-- # import Implicit.Terms
-- # import Implicit.Substitutions
-- # import Implicit.Equality
-- # import Implicit.Equivalence
-- # import Implicit.EqualityLemmas
-- # import EqKind
-- # import TmVar

abstract

  tm⁻EqTmEq (k : EqKind)
            (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
            (t⁻¹ : Tm⁻ tv¹ t¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
            (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> t¹ ≈⊢ t²
  tm⁻EqTmEq (var⁻Cong eqv)       = varCong eqv
  tm⁻EqTmEq (λ⁻Cong eqt)         = λCong (tm⁻EqTmEq eqt)
  tm⁻EqTmEq (app⁻Cong eqt₁ eqt₂) =
    appCong (tm⁻EqTmEq eqt₁) (tm⁻EqTmEq eqt₂)
  tm⁻EqTmEq (castTm⁻Eqˡ eqt eqt⁻) = symTm eqt `transTm` tm⁻EqTmEq eqt⁻
  tm⁻EqTmEq (castTm⁻Eqʳ eqt eqt⁻) = tm⁻EqTmEq eqt⁻ `transTm` eqt
  tm⁻EqTmEq (symTm⁻βη eqt⁻)       = symTm $ tm⁻EqTmEq eqt⁻
  tm⁻EqTmEq (transTm⁻βη eqt⁻¹² eqt⁻²³) =
    transTm (tm⁻EqTmEq eqt⁻¹²) (tm⁻EqTmEq eqt⁻²³)
  tm⁻EqTmEq (β⁻ t⁻₁ t⁻₂) = β _ _
  tm⁻EqTmEq (η⁻ t⁻)      = η _

  tm⁻EqTyEq (k : EqKind)
            (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
            (t⁻¹ : Tm⁻ tv¹ t¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
            (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> τ¹ ≛ τ²
  tm⁻EqTyEq eqt⁻ = tmEqTyEq $ tm⁻EqTmEq eqt⁻

  tm⁻EqCtxtEq (k : EqKind)
              (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
              (t⁻¹ : Tm⁻ tv¹ t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
              (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> Γ¹ ≜ Γ²
  tm⁻EqCtxtEq eqt⁻ = tmEqCtxtEq $ tm⁻EqTmEq eqt⁻

  tm⁻ToTmCong (k : EqKind)
              (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
              (t⁻¹ : Tm⁻ tv¹ t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
              (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> tm⁻ToTm t⁻¹ ≈⊢ tm⁻ToTm t⁻²
  tm⁻ToTmCong eqt =
    betweenTm (tm⁻ToTmEq _) (tm⁻ToTmEq _) $
      tm⁻EqTmEq eqt

  tm⁻ToVarCongˢ
    (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (t⁻¹ : Tm⁻ Var t¹)
    (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (t⁻² : Tm⁻ Var t²)
    : Tm⁻Eq structural t⁻¹ t⁻² -> tm⁻ToVar t⁻¹ ≈∋ tm⁻ToVar t⁻²
  tm⁻ToVarCongˢ (var⁻Cong eqv) = eqv
  tm⁻ToVarCongˢ (castTm⁻Eqˡ _ eqt⁻) =
    castVarEqˡ _ $ tm⁻ToVarCongˢ eqt⁻
  tm⁻ToVarCongˢ (castTm⁻Eqʳ _ eqt⁻) =
    castVarEqʳ _ $ tm⁻ToVarCongˢ eqt⁻

  ty⁻EqTyEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (τ⁻¹ : Ty⁻ τ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (τ⁻² : Ty⁻ τ²)
    : τ⁻¹ ≛⁻ τ⁻² -> τ¹ ≛ τ²
  ty⁻EqTyEq (✶⁻Cong eqΓ)         = ✶Cong eqΓ
  ty⁻EqTyEq (Π⁻Cong eqτ⁻₁ eqτ⁻₂) =
    ΠCong (ty⁻EqTyEq eqτ⁻₁) (ty⁻EqTyEq eqτ⁻₂)
  ty⁻EqTyEq (El⁻Cong eqt⁻)        = ElCong (tm⁻EqTmEq eqt⁻)
  ty⁻EqTyEq (castTy⁻Eqˡ eqτ eqτ⁻) = symTy eqτ `transTy` ty⁻EqTyEq eqτ⁻
  ty⁻EqTyEq (castTy⁻Eqʳ eqτ eqτ⁻) = ty⁻EqTyEq eqτ⁻ `transTy` eqτ

  ty⁻ToTyCong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (τ⁻¹ : Ty⁻ τ¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (τ⁻² : Ty⁻ τ²)
    : τ⁻¹ ≛⁻ τ⁻² -> ty⁻ToTy τ⁻¹ ≛ ty⁻ToTy τ⁻²
  ty⁻ToTyCong eqτ⁻ =
    betweenTy (ty⁻ToTyEq _) (ty⁻ToTyEq _) $
      ty⁻EqTyEq eqτ⁻

  from⁻SubstCommutes
    (tv : TmVar) (Γ, Δ : Ctxt) (τ : Ty Γ) (ρ : Subst Γ Δ)
    :  (v : Γ ∋ τ) -> (ρ⁻ : Subst⁻ tv ρ)
    -> var v /⊢ subst⁻ToSubst ρ⁻ ≈⊢ tm⁻ToTm (v /∋⁻ ρ⁻)
  from⁻SubstCommutes v (ρ ∷≡⇒⁻ eqρ) =
    castTmEqʳ _ $ from⁻SubstCommutes v ρ
  from⁻SubstCommutes (vz σ) (ρ ▷⁻ t⁻) =
    substVzSub' σ (subst⁻ToSubst ρ) (tm⁻ToTm t⁻ ∷≡⊢ _)
      `transTm`
    castTmCong _ _ (reflTm _)
  from⁻SubstCommutes (vs v σ) (ρ ▷⁻ t⁻) =
    substVsSub' v σ (subst⁻ToSubst ρ) (tm⁻ToTm t⁻ ∷≡⊢ _)
      `transTm`
    castTmEqʳ _ (from⁻SubstCommutes v ρ)
  from⁻SubstCommutes (v ∷≡∋ eqτ) ρ =
    substCastVar v eqτ (subst⁻ToSubst ρ)
      `transTm`
    from⁻SubstCommutes v ρ
      `transTm`
    symTm (tm⁻ToTmCong $ substCastVar⁻ {structural} v eqτ ρ)

  subst⁻ToSubstCong (k : EqKind)
                    (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt)
                    (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
                    (tv² : TmVar) (Γ², Δ² : Ctxt)
                    (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    : Subst⁻Eq k ρ⁻¹ ρ⁻² -> subst⁻ToSubst ρ⁻¹ ≘ subst⁻ToSubst ρ⁻²
  subst⁻ToSubstCong (extEq⁻ eqΓ eqΔ proof) = extEqToSubstEq $
    extEq eqΓ eqΔ (\_ v¹ _ v² eqv ->
                      betweenTm (from⁻SubstCommutes v¹ _)
                                (from⁻SubstCommutes v² _) $
                        tm⁻ToTmCong $ proof _ v¹ _ v² eqv)

  subst⁻EqSubstEq (k : EqKind)
                  (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt)
                  (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
                  (tv² : TmVar) (Γ², Δ² : Ctxt)
                  (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    : Subst⁻Eq k ρ⁻¹ ρ⁻² -> ρ¹ ≘ ρ²
  subst⁻EqSubstEq eqρ⁻ =
    betweenSubst (symSubst (subst⁻ToSubstEq _))
                 (symSubst (subst⁻ToSubstEq _)) $
      subst⁻ToSubstCong eqρ⁻

Implicit-Soundness : Type
Implicit-Soundness = time-plugin(Implicit-Soundness, Set)
