------------------------------------------------------------------------
-- Data type used to index things which may contain either plain
-- variables or terms
------------------------------------------------------------------------

-- # import Eq

infixl 50 ⊔

-- The data type is defined in a somewhat awkward way to make sure
-- that Var is structurally smaller than Term. If we could use defined
-- patterns, then we could pattern match on Term instead of Term' Var
-- refl, and then the damage would be negligible.

data TmVar : Set where
  Var   : TmVar
  Term' : (tv : TmVar) -> tv ≡ Var -> TmVar

Term : TmVar
Term = Term' Var refl

(⊔) : TmVar -> TmVar -> TmVar
Term' Var refl ⊔ tv = Term
Var            ⊔ tv = tv

------------------------------------------------------------------------
-- Some lemmas

abstract

  VarRightUnit : (tv : TmVar) -> tv ⊔ Var ≡ tv
  VarRightUnit Var              = refl
  VarRightUnit (Term' Var refl) = refl

  TermRightZero : (tv : TmVar) -> tv ⊔ Term ≡ Term
  TermRightZero Var              = refl
  TermRightZero (Term' Var refl) = refl

  idempotent⊔ : (tv : TmVar) -> tv ⊔ tv ≡ tv
  idempotent⊔ Var              = refl
  idempotent⊔ (Term' Var refl) = refl
